//*******************************************************
//       Copyright (c) 2003 Gena A. Petrov
//*******************************************************
unit ResourceStringsManage;

interface

uses
  Classes, Forms;

type
  PString = ^String;
  PResStrRec = ^TResStrRec;
  TResStrRec = record
    StrName: String;
    pStrValue: PString;
  end;
  TResStrRecList = array of PResStrRec;

  TResStrItem = class(TCollectionItem)
  private
    FStringValue: String;
    FStringName: String;
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
  published
    property StringName: String read FStringName write FStringName;
    property StringValue: String read FStringValue write FStringValue;
  end;

  TResStrItems = class(TCollection)
  private
    FControl: TComponent;
  protected
    procedure SetItem(Index: Integer; Value: TResStrItem);
    function GetItem(Index: Integer): TResStrItem;
    function Insert(Index: Integer): TResStrItem;
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(Control: TComponent);

    procedure Assign(Source: TPersistent); override;

    function Add: TResStrItem;
    property Items[Index: Integer]: TResStrItem read GetItem write SetItem; default;
  end;

  TResourceStrings = class(TComponent)
  private
    FRSItems: TResStrItems;

    procedure SetResourceStrings(Value: TResStrItems);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property ResourceStrings: TResStrItems read FRSItems write SetResourceStrings;
  end;

  TResourceStringsManager = class(TObject)
  private
    FRSRAItems: TResStrRecList;
    FOtnCount,    
    FCount: Integer;
    procedure Update(Value: TResStrItems); overload;
  public
    constructor Create(AOwner: TComponent);
    destructor Destroy; override;
    procedure SetStringValue(const StrName: String; var StrValue: String);
    procedure Update; overload;
  end;

var
  ResourceStringsManager: TResourceStringsManager;

procedure Register;

implementation

uses
  Windows,
  SysUtils;

procedure Register;
begin
  RegisterComponents('FormsConstructor', [TResourceStrings]);
end;

procedure SortRSRAItems(RSRL: TResStrRecList);
  procedure QuickSortExt(L,R: Integer);
  var
    L_,
    R_: Integer;
    w: Pointer;
    Str: String;
  begin
    repeat
      L_ := L;
      R_ := R;
      if L >= R then exit;
      Str := RSRL[(L + R) shr 1]^.StrName;
      repeat
        while CompareText(RSRL[L_]^.StrName,Str) < 0 do
          Inc(L_);
        while CompareText(RSRL[R_]^.StrName, Str) > 0 do
          Dec(R_);
        if L_ <= R_ then
        begin
          w:= RSRL[R_];
          RSRL[R_]:= RSRL[L_];
          RSRL[L_]:= w;
          Inc(L_);
          Dec(R_);
        end;
      until L_ > R_;
      if L < R_ then QuickSortExt(L, R_);
      L := L_;
    until L_ >= R;
  end;

begin
  if RSRL = nil then exit;

  QuickSortExt(0,Length(RSRL)-1);
end;

function QuickFindStrName(StrName: String; RSRL: TResStrRecList): Integer;
var
  i,
  L,
  R,
  M: Integer;
begin
  Result:= -1;

  L:= 0;
  R:= Length(RSRL)-1;

  if R = -1 then exit;

  repeat
    M := (R+L) div 2;

    if RSRL[M] <> nil then
    begin
      i := CompareText(RSRL[M]^.StrName,StrName);
      if i < 0 then
        L := M+1
      else
      begin
        if i = 0 then
        begin
          Result:= M;
          exit;
        end
        else
          R := M-1;
      end;
    end;
  until L > R;
end;

constructor TResStrItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);

  FStringName := '';
end;

procedure TResStrItem.Assign(Source: TPersistent);
begin
  if Source is TResStrItem then
  begin
    StringName:= TResStrItem(Source).StringName;
    StringValue:= TResStrItem(Source).StringValue;
  end
  else
    inherited Assign(Source);
end;

function TResStrItem.GetDisplayName: string;
begin
  Result := StringName;
  if Result = '' then Result := inherited GetDisplayName;
end;

constructor TResStrItems.Create(Control: TComponent);
begin
  inherited Create(TResStrItem);
  FControl := Control;
end;

function TResStrItems.Add: TResStrItem;
begin
  Result := TResStrItem.Create(Self);//TResStrItem(inherited Add);
end;

procedure TResStrItems.Assign(Source: TPersistent);
var
  I: Integer;
begin
  if Source is TResStrItems then
  begin

    BeginUpdate;
    try
      Clear;
      for I := 0 to TResStrItems(Source).Count - 1 do
        Add.Assign(TResStrItems(Source).Items[I]);
    finally
      EndUpdate;
    end;
    Exit;
  end;
  inherited Assign(Source);
end;

function TResStrItems.GetItem(Index: Integer): TResStrItem;
begin
  Result := TResStrItem(inherited GetItem(Index));
end;


function TResStrItems.Insert(Index: Integer): TResStrItem;
begin
  Result := Add;
  Result.Index := Index;
end;

procedure TResStrItems.SetItem(Index: Integer; Value: TResStrItem);
begin
  inherited SetItem(Index, Value);
end;

function TResStrItems.GetOwner: TPersistent;
begin
  Result := FControl;
end;

procedure TResStrItems.Update(Item: TCollectionItem);
begin
  inherited;
  if not (csDesigning in TResourceStrings(GetOwner).ComponentState) then
    ResourceStringsManager.Update(Self);
end;


constructor TResourceStrings.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FRSItems:= TResStrItems.Create(Self);
end;

destructor TResourceStrings.Destroy;
begin
  FRSItems.Free;

  inherited Destroy;
end;

procedure TResourceStrings.SetResourceStrings(Value: TResStrItems);
begin
  FRSItems:= Value;
  ResourceStringsManager.Update(Value);
end;


constructor TResourceStringsManager.Create(AOwner: TComponent);
begin
  FRSRAItems := nil;
  FCount := 0;
  FOtnCount := 0;
end;

destructor TResourceStringsManager.Destroy;
begin
  SetLength(FRSRAItems,0);
end;

procedure TResourceStringsManager.Update(Value: TResStrItems);
var
  OtnCount,
  l,
  i: Integer;

  A: array of integer;
  AotnCount,
  ACount: Integer;
begin
  AotnCount:= 0;
  ACount:= 0;
  OtnCount:= 0;

  for i:= 0 to Value.Count-1 do
  begin
    l := QuickFindStrName( Value.Items[i].StringName, FRSRAItems);
    if l <> -1 then
    begin
      FRSRAItems[l].pStrValue^ := Value.Items[i].StringValue;
      //SetLength(Value.Items[i].FStringValue,0);
      //Value.Items[i].StringValue := String(FRSRAItems[l].pStrValue);
    end
    else
    begin
      if AotnCount < 1 then
      begin
        AotnCount:= 5;
        SetLength(A, ACount+AotnCount);
      end;
      A[ACount]:= i;
      Inc(ACount);
      Dec(AotnCount);
    end;
  end;

  
  for i:= 0 to ACount-1 do
  begin
    if OtnCount < 1 then
    begin
      OtnCount:= 5;
      SetLength(FRSRAItems, FCount + OtnCount);
    end;

    New(FRSRAItems[FCount]);
    FRSRAItems[FCount]^.StrName := Value.Items[i].FStringName;
    FRSRAItems[FCount]^.pStrValue := @Value.Items[i].FStringValue;

    Inc(FCount);
    Dec(OtnCount);
  end;

  SetLength(A,0);
  SetLength(FRSRAItems, FCount);
  SortRSRAItems(FRSRAItems);
end;

procedure TResourceStringsManager.Update;
begin
  SetLength(FRSRAItems, FCount);
  SortRSRAItems(FRSRAItems);
end;

procedure TResourceStringsManager.SetStringValue(const StrName: String; var StrValue: String);
begin
  if FOtnCount < 1 then
  begin
    FOtnCount := 5;
    SetLength(FRSRAItems, FCount + FOtnCount);
  end;

  New(FRSRAItems[FCount]);
  FRSRAItems[FCount]^.StrName := StrName;
  FRSRAItems[FCount]^.pStrValue := @StrValue;
  Inc(FCount);
  Dec(FOtnCount);
end;

initialization
  ResourceStringsManager:= TResourceStringsManager.Create(nil);

finalization
  ResourceStringsManager.Free;

end.
